from __future__ import unicode_literals, absolute_import


def is_win32():
    """
    Check os platform is win32

    Returns:
        bool: True if is win32 else False
    """
    import sys
    return "win32" in sys.platform


def is_linux():
    """
    Check os platform is linux

    Returns:
        bool: True if is linux else False
    """
    import sys
    return "linux" in sys.platform


def is_python2():
    """
    Check current interpreter is python2

    Returns:
        bool: True if is python2 else False
    """
    import sys
    return sys.version_info.major == 2


def is_python3():
    """
    Check current interpreter is python3

    Returns:
        bool: True if is python3 else False
    """
    import sys
    return sys.version_info.major == 3


def set_unicode():
    """
    Set default encoding to utf8
    """
    if is_python3():
        return

    import sys
    reload(sys)
    sys.setdefaultencoding("utf8")


def execute(command, callback=None, timeout=0):
    '''
    Execute a system command return status and output

    Args:
        command (TYPE): execute command command must not bash command
        callback (None, optional): callback function per output line

    Returns:
        string: console output if callback is None
        int: execute exit code
    '''
    import time
    import subprocess
    if not is_win32():
        cmd = ["/bin/sh", "-c", command]
        shell = False
    else:
        cmd = command
        shell = True
    try:
        pipe = subprocess.Popen(
            cmd, shell=shell, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    except Exception as e:
        return e, -1

    if not callback:
        res = pipe.stdout.read()
        pipe.communicate()
        if type(res) == bytes:
            line = ""
            for char in res:
                try:
                    char = chr(char)
                    line += char
                except Exception:
                    continue
            res = line
        return (res, pipe.returncode)

    start_time = time.time()
    line = ""
    while True:
        char = pipe.stdout.read(1)
        if not char:
            break
        if type(char) == bytes:
            try:
                char = bytes.decode(char)
            except Exception:
                continue
        if char not in ["\n", "\r"]:
            line += char
            continue
        callback(line)
        if timeout and (time.time() - start_time) > timeout:
            kill(pipe.pid)
            break
            # return (None, 9)
        line = ""
    pipe.communicate()
    return (None, pipe.returncode)


def kill(pid):
    '''
     Kill progress with pid

    Args:
        pid (int): progress id
    '''
    import psutil

    process = psutil.Process(pid)
    for proc in process.children(recursive=True):
        try:
            proc.kill()
        except Exception:
            continue
    process.kill()


def kill_command(command):
    '''
    Kill progress with command

    Args:
        command (string): run command
    '''
    if not command:
        return

    import re

    if is_win32():
        tup = command.split(" ", 1)
        caption = tup[0]
        if len(tup) > 1:
            run_para = tup[1]
        cmd = '''wmic process where name="{}" get commandline,processid'''.format(caption)
        result, code = execute(cmd)
        results = result.splitlines()
        for result in results:
            match = re.search(r"\w+ +(?P<para>.+) +(?P<pid>\d+)", result)
            if not match:
                continue
            para = match.group("para").strip()
            if para != run_para:
                continue
            pid = match.group("pid")
            kill(pid)
        return

    elif is_linux():
        cmd = '''ps -ef | grep "{}" | grep -v grep'''.format(command)
        result, code = execute(cmd)
        results = result.splitlines()
        for result in results:
            result = re.sub(" +", " ", result)
            match = re.search(r"\w+ (?P<pid>\d+) .+:\d\d (?P<cmd>.+)", result)
            if not match:
                continue
            cmd = match.group("cmd")
            if cmd not in [command, '/bin/sh -c ' + command]:
                continue
            pid = match.group("pid")
            kill(pid)
        pass


def readable(path):
    '''
    Check if a given path is readable by the current user.

    Args:
        path (string): local system path

    Returns:
        bool: True if readable else False
    '''

    import os
    if os.access(path, os.F_OK) and os.access(path, os.R_OK):
        return True

    return False


def writeable(path, check_parent=True):
    '''
    Check if a given path is writeable by the current user.

    Args:
        * path (string): local system path
        * check_parent (bool, optional): If the path to check does not exist, check for the ability to write to the parent directory instead

    Returns:
        * bool: True if writeable else False
    '''
    import os
    if os.access(path, os.F_OK) and os.access(path, os.W_OK):
        # The path exists and is writeable
        return True

    if os.access(path, os.F_OK) and not os.access(path, os.W_OK):
        # The path exists and is not writeable
        return False

    # The path does not exists or is not writeable

    if check_parent is False:
        # We're not allowed to check the parent directory of the provided path
        return False

    # Lets get the parent directory of the provided path
    parent_dir = os.path.dirname(path)

    if not os.access(parent_dir, os.F_OK):
        # Parent directory does not exit
        return False

    # Finally, return if we're allowed to write in the parent directory of the
    # provided path
    return os.access(parent_dir, os.W_OK)


def __getch_win32():
    import msvcrt
    return msvcrt.getch()


def __getch_linux():
    import termios
    import sys
    import os

    fd = sys.stdin.fileno()

    old_ttyinfo = termios.tcgetattr(fd)
    new_ttyinfo = old_ttyinfo[:]

    new_ttyinfo[3] &= ~termios.ICANON
    new_ttyinfo[3] &= ~termios.ECHO

    try:
        termios.tcsetattr(fd, termios.TCSANOW, new_ttyinfo)
        return os.read(fd, 7)
    except Exception:
        return chr(3)
    finally:
        termios.tcsetattr(fd, termios.TCSANOW, old_ttyinfo)


def getch():
    """
    Get a char pressed on keyboard without press Enter

    Returns:
        char(s): return pressed key
    """
    if is_win32():
        ch = __getch_win32()
    elif is_linux():
        ch = __getch_linux()
    else:
        ch = None

    if is_python2():
        return unicode(ch)
    elif is_python3():
        return ch.decode('utf8')


def clear():
    """
    Clear console screen
    """
    import os
    if is_win32():
        os.system("cls")
    elif is_linux():
        os.system("clear")



from __future__ import unicode_literals, absolute_import


def local_ip():
    '''get local ip address for IP V4

    Returns:
        * string: current machine ip

    TODO:
        ip v6 not implement
    '''

    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 0))
        ip = s.getsockname()[0]
    except Exception:
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip


def html(url, **kwargs):
    '''get html string object by url

    User-Agent :
        default "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:52.0) Gecko/20100101 Firefox/52.0"
        modify User-Agent put to headers in kwargs

    Args:
        * url (string): requested http url
        * kwargs
        * timeout : set get timeout default 60
        * headers : set http headers have User-Agent
        * params : set http parameters
        * retry : set retry count default 0
        * encoding : page encoding default utf8
        * method : set http method default get
        * json : if True return json else return string

    Returns:
        string: if request correct else None
    '''

    from dandan import value
    import requests
    # import traceback

    kwargs = value.AttrDict(kwargs)
    timeout = kwargs.timeout or 60
    headers = kwargs.headers or value.AttrDict()
    params = kwargs.params or value.AttrDict()
    method = kwargs.method or 'get'
    retry = kwargs.retry or 0
    encoding = kwargs.encoding or "utf8"
    json = kwargs.json or False

    user_agent = "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:52.0) Gecko/20100101 Firefox/52.0"
    headers["User-Agent"] = headers["User-Agent"] or user_agent

    for _ in range(0, retry + 1):
        try:
            if method.lower() == 'post':
                res = requests.post(url, data=params.dict(), headers=headers, timeout=timeout)
            else:
                res = requests.get(url, params=params, headers=headers, timeout=timeout)
            res.encoding = encoding
            if json:
                return res.json()
            return res.text
        except Exception:
            continue

    return None


def soup(url, **kwargs):
    r'''get BeautifulSoup object by url

    Args:
        * url (string): requested http url
        * \*\*kwargs: same as :py:func:`html`

    Returns:
        BeautifulSoup: if request correct else None
    '''

    from bs4 import BeautifulSoup as bs
    text = html(url, **kwargs) or ""
    return bs(text, "html.parser")


def json(url, **kwargs):
    r'''get json object by url familiar html if http return json or None

    Args:
        * url (string): requested http url
        * \*\*kwargs: same as :py:func:`html`

    Returns:
        json: if request correct else None
    '''
    return html(url, json=True, **kwargs)


def whois(ip):
    """
    Get whois infommation (developing)

    Args:
        * ip (string): request for whois

    Returns:
        * AttrDict: if get whois currect else None
    """
    import value

    result = value.AttrDict()
    result.ip = ip

    api = "http://freegeoip.net/json/{}".format(ip)
    res = json(api, timeout=5)
    if not res:
        return result

    res = value.AttrDict(res)

    result.region = res.region_name
    return result



# encoding=utf-8
from __future__ import absolute_import

import os
import time
import requests


def download(url, filename, callback=None, force=False, headers={}, check_length=False):
    """
    Download http file to filename

    Args:
        * url (string): http url for download
        * filename (string): local system filename to save file
        * callback (function, optional): callback function when file downloading
        * force (bool, optional): redownload if filename exists default False
        * headers (dict, optional): http headers
        * check_length (bool, optional): Check length for downloaded file and http headers content-length

    Returns:
        * bool: success status
        * string: file length if correct else error information
        * int : spend time for download period
    """
    start = time.time()
    if not force and os.path.exists(filename) and os.path.getsize(filename) > 0:
        return (True, filename, time.time() - start)
    filepath = os.path.dirname(os.path.abspath(filename))
    if not os.path.exists(filepath):
        os.makedirs(filepath)
    temp_filename = filename + ".tmp"
    try:
        res = requests.get(url, headers=headers, stream=True, timeout=60)
        if res.status_code != 200:
            return (False, "HTTP code {0}".format(res.status_code), time.time() - start)
        f = open(temp_filename, "wb")
        length = int(res.headers.get("content-length", 0))
        buflen = 4096
        count = 0
        for chunk in res.iter_content(buflen):
            count += len(chunk)
            f.write(chunk)
            if callback:
                callback(count, length, time.time() - start)
        f.close()
        if check_length and count != length:
            return (False, "content-length not equal", time.time() - start)
        os.rename(temp_filename, filename)
        return (True, filename, time.time() - start)
    except Exception as e:
        return (False, str(e), time.time() - start)
    finally:
        try:
            f.close()
            res.close()
        except Exception:
            pass


def upload(filename, url, callback=None, **kwargs):
    """
    Upload filename to url

    Args:
        * filename (string): local system filename to upload
        * url (string): http url
        * callback (function, optional): callback function when file downloading
        * kwargs: parameters for http

    Returns:
        * bool: success status
        * string: url return content if correct else error information
        * int : spend time for download period
    """
    class IterableToFileAdapter(object):
        def __init__(self, iterable):
            self.iterator = iter(iterable)
            self.length = iterable.total

        def read(self, size=-1):
            return next(self.iterator, b'')

        def __len__(self):
            return self.length

    import poster
    from collections import OrderedDict

    start = time.time()
    if not os.path.exists(filename):
        return (False, "File not exists.", time.time() - start)

    data = OrderedDict(**kwargs)
    files = None
    headers = None

    def real_callback(p, current, total):
        if not callback:
            return
        callback(current, total, time.time() - start)

    if not callback:
        files = {'files': open(filename, 'rb')}
    else:
        data["files"] = open(filename, 'rb')
        datagen, headers = poster.encode.multipart_encode(
            params=data, cb=real_callback)
        data = IterableToFileAdapter(datagen)

    try:
        res = requests.post(url,
                            data=data,
                            headers=headers,
                            files=files,
                            timeout=60)
        success = False
        if res.status_code == 200:
            success = True
        return (success, res.text.strip(), time.time() - start)
    except Exception as e:
        return (False, e, time.time() - start)



from __future__ import absolute_import


def interrupt(func):
    """Decorator for Interrupt function

    Args:
        func (function): Description
    """
    import functools

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return

    return wrapper



# from __future__ import absolute_import
import os
import sys
from setuptools import setup

NAME = "dandan"
INSTALL_REQUIRES = [
    "six",
    "requests",
    "beautifulsoup4",
]

base_dir = os.path.dirname(os.path.abspath(__file__))
if base_dir not in sys.path:
    sys.path.insert(0, base_dir)
try:
    import dandan
    version = dandan.__version__
except Exception:
    print("import project error")
    exit(0)

readme = os.path.join(os.path.dirname(os.path.abspath(__file__)), "README")

setup(
    name=NAME,
    version=dandan.__version__,
    description="Several convenient tools for python programming",
    long_description=open(readme).read(),
    packages=[NAME],
    install_requires=INSTALL_REQUIRES,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    author="ccyg",
    author_email="kjp1314@163.com",
    url="https://github.com/ccyg/dandan",
    license="MIT",
    include_package_data=True,
    zip_safe=True,
    platforms="any",
)


# python setup.py sdist bdist_wheel --universal
# twine upload dist/*
# pip2 install dandan --upgrade -i https://pypi.python.org/simple
# pip3 install dandan --upgrade -i https://pypi.python.org/simple



"""
**dandan package**

Several convenient tools for python programming


**Events**

- 2020-10-10 [0.7.4] update info
- 2020-10-10 [0.7.3] remove psutil in dependencies
- 2018-10-13 [0.7.2] fix bug for AttrDict call dict.update
- 2018-10-11 [0.7.1] optimize performance for AttrDict and fix bug
- 2018-10-11 [0.7.0] optimize performance for AttrDict
- 2018-05-21 [0.6.0] add interrupt decorator
- 2018-05-16 [0.5.8] fix bug for AttrDict member functions
- 2018-05-12 [0.5.7] improve code robustness
- 2018-05-11 [0.5.6] modify logger roll suffix is "%Y-%m-%d.log"
- 2018-05-07 [0.5.5] fix bug for md5 and sha1
- 2018-05-03 [0.5.4] fix bug for logger file utf8 encoding
- 2018-04-28 [0.5.3] fix bug for logger file
- 2018-03-23 [0.5.2] add default logger name as 'dandan'
- 2017-12-13 [0.5.1] fix bug
- 2017-12-13 [0.5.0] add system.kill and execute timeout
- 2017-12-12 [0.4.2] fix bug for system.execute in callback mode
- 2017-12-11 [0.4.1] fix bug for system.execute return result with str
- 2017-12-10 [0.4.0] add getLogger method in logger
- 2017-11-29 [0.3.3] fix bug for put_json in python3
- 2017-11-29 [0.3.2] add indent for dandan.value.put_json
- 2017-11-19 [0.3.1] fix bug in dandan.value.length when given string length is zero
- 2017-11-19 [0.3.0] add function dandan.value.length
- 2017-11-17 [0.2.3] update document for project enhance AttrDict class
- 2017-11-15 [0.2.2] update document for project
- 2017-10-14 move to another github project
- 2017-06-25 add function system.clear
- 2017-06-23 add function system.getch
- 2017-06-23 Support python3
- 2017-06-23 Add TestCase
"""

from __future__ import absolute_import

from dandan import error
from dandan import query
from dandan import system
from dandan import traffic
from dandan import value
from dandan import utils
from dandan import logger

__all__ = [
    "error",
    "query",
    "system",
    "traffic",
    "value",
    "utils",
    "logger"
]

__version__ = ".".join(
    [str(var) for var in
        [
            0,
            7,
            4
    ]
    ])



# encoding=utf-8
from __future__ import absolute_import

import json


class AttrDict(dict):

    """
    Use dict key as attr

    **examples**

    .. code-block:: python
        :linenos:

        data = Attrdict()

        # after two lines are equal statement
        data.key1 = 1
        data["key1"] = 1


        # defaut value also is Attrdict after line is allowed
        data.key2.key.key.key = 5

        # if plus or minus a number the default value is 0
        data.key3 += 5
        assest(data.key2 == 5)

    enjoys!!!
    """

    DICT_ATTRS = set(dir(dict))

    def __init__(self, dic={}, json_string=None, *args, **kwargs):

        self.update(dict(args))
        self.update(kwargs)
        self.update(dic)
        if json_string:
            dic = json.loads(json_string)
            self.update(dic)

        for key, value in self.items():
            if type(value) in (dict, list):
                self[key] = self.__updatevalue__(value)

    def __getattr__(self, key):
        if key in self.DICT_ATTRS:
            return dict.__getattr__(self, key)
        if key not in self:
            self[key] = AttrDict()
        return self[key]

    def __setattr__(self, key, value):
        if key in self.DICT_ATTRS:
            return dict.__setattr__(self, key, value)
        self[key] = value

    def __delattr__(self, key):
        if key in self.DICT_ATTRS:
            return dict.__delattr__(self, key)
        del self[key]

    def __delitem__(self, key):
        if key not in self:
            return
        dict.__delitem__(self, key)

    def __getitem__(self, key):
        if key in self.DICT_ATTRS:
            return dict.__getitem__(self, key)
        elif key not in self:
            self[key] = type(self)()
        return dict.__getitem__(self, key)

    def __updatevalue__(self, value):
        if type(value) is dict:
            value = type(self)(value)
        elif type(value) is list:
            for index, item in enumerate(value):
                if type(item) is dict:
                    value[index] = type(self)(item)
        return value

    def __setitem__(self, key, value):
        if key in self.DICT_ATTRS:
            return dict.__setitem__(self, key, value)
        dict.__setitem__(self, key, self.__updatevalue__(value))

    def dict(self):
        """
        return dict object of current instance

        Returns:
            * dict: convert from current instance
        """
        res = {}
        for var in self:
            value = self[var]
            if isinstance(value, type(self)):
                value = value.dict()
            if isinstance(value, list):
                values = []
                for item in value:
                    if isinstance(item, type(self)):
                        values.append(item.dict())
                    else:
                        values.append(item)
                value = values
            res[var] = value
        return res

    def __getstate__(self):
        return self.dict()

    def __setstate__(self, state):
        pass

    def __add__(self, other):
        if not self:
            return other
        raise ValueError("Dict not empty.")

    def __sub__(self, other):
        if not self:
            return -other
        raise ValueError("Dict not empty.")

    def __str__(self):
        return super(AttrDict, self).__str__()

    def __unicode__(self):
        return super(AttrDict, self).__str__()

    def __copy__(self):
        import copy
        return type(self)(copy.copy(self.dict()))

    def __deepcopy__(self, memo):
        from copy import deepcopy
        return type(self)(deepcopy(self.dict()))

    def __reduce__(self):
        return super(AttrDict, self).__reduce__()

# function


def get_pickle(filename):
    """
    Load file as pickle object

    Args:
        * filename (string): local system filename

    Returns:
        * object: if file is pickled data else None
    """
    import os
    import pickle
    if not os.path.exists(filename):
        return None
    try:
        with open(filename, "rb") as f:
            data = pickle.load(f)
            return data
    except Exception:
        return None


def put_pickle(data, filename):
    """
    Save object as pickle to filename

    Args:
        * data (object): any can pickled object
        * filename (string): local system filename
    """
    import os
    import pickle
    filename = os.path.abspath(filename)
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    with open(filename, "wb") as f:
        pickle.dump(data, f)


def get_json(filename):
    """
    Load file as pickle object

    Args:
        * filename (string): local system filename

    Returns:
        * object: if file is json data else None
    """
    import os
    import json
    if not os.path.exists(filename):
        return None
    try:
        with open(filename, "rb") as f:
            data = json.loads(f.read())
            return data
    except Exception:
        return None


def put_json(data, filename, indent=None):
    """
    Save object as json string to filename

    Args:
        * data (object): any can jsoned object
        * filename (string): local system filename
        * indent (None, optional): json indent width default None
    """
    import os
    import json
    filename = os.path.abspath(filename)
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    with open(filename, "w") as f:
        f.write(json.dumps(data, indent=indent))


def md5(data=None, filename=None, encoding="utf8"):
    '''
    get data or file md5 checksum

    Args:
        * data (string, optional): string
        * filename (string, optional): local system filename

    Returns:
        * string: md5 checksum
    '''
    import hashlib
    m = hashlib.md5()
    if data:
        m.update(data.encode(encoding=encoding))
    if filename:
        with open(filename, "rb") as f:
            while True:
                block = f.read(16 * 1024)
                if not block:
                    break
                m.update(block)
    return m.hexdigest()


def sha1(data=None, filename=None, encoding="utf8"):
    '''
    get data or file sha1 checksum

    Args:
        * data (string, optional): string
        * filename (string, optional): local system filename

    Returns:
        * string: sha1 checksum
    '''
    import hashlib
    m = hashlib.sha1()
    if data:
        m.update(data.encode(encoding=encoding))
    if filename:
        with open(filename, "rb") as f:
            while True:
                block = f.read(16 * 1024)
                if not block:
                    break
                m.update(block)
    return m.hexdigest()


def is_number(number):
    '''
    Test parameter is number True or False

    Args:
        * number (TYPE): any object

    Returns:
        * bool: True if number is float value or False
    '''
    try:
        float(number)
        return True
    except Exception:
        return False


def number(num):
    '''
    convert parameter to float or None

    Args:
        * num (TYPE): any object

    Returns:
        * float: num float value or None
    '''
    try:
        return float(num)
    except Exception:
        return None


def length(string):
    '''
    Get true size of string or char, char might be 1 or 2, string accumulate all of char

    **example**

    .. code-block:: python
        :linenos:

        dandan.value.length("test string") == 11
        dandan.value.length("测试字符串") == 10
        dandan.value.length("测试字符串 test string") == 22

    Args:
        string (string or char): requested

    Returns:
        int: size of string or char
    '''
    import six
    if six.PY2:
        string = unicode(string)

    if len(string) == 0:
        return 0
    if len(string) > 1:
        accumulate = 0
        for char in string:
            accumulate += length(char)
        return accumulate

    import unicodedata
    if unicodedata.east_asian_width(string) in {'F', 'W', 'A'}:
        return 2
    else:
        return 1



# encoding=utf-8
from __future__ import absolute_import

import logging
import logging.config

from dandan import value

def getLogger(name="dandan", level=logging.DEBUG, filename=None, backup_count=10):
    '''
    Get logger for convenient method

    Args:
        * name (string): logger name, default as 'dandan'
        * level (logger level, optional): level of this logger such as DEBUG, INFO, WARNING, ERROR, FATAL
        * filename (string, optional): filename for timerotedlogger
        * backup_count (int, optional): file backup count, if file count larger than count, then oldest file will be deleted.

    Returns:
        * logger: the logger named name
    '''

    logger = logging.getLogger(name)
    if len(logger.handlers) > 0:
        return logger
    config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'verbose': {
                'format': '[%(asctime)s] [%(module)s] [%(lineno)d] [%(levelname)s] | %(message)s'
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'verbose',
                "level": "DEBUG",
            },
            'null': {
                'class': 'logging.NullHandler',
            },
        },
        'loggers': {
            name: {
                'handlers': ['console', ],
                'level': level,
                'propagate': True,
            },
        },
    }
    config = value.AttrDict(config)
    if filename:
        config.handlers.file = {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'formatter': 'verbose',
            'filename': filename,
            'when': "MIDNIGHT",
            "level": "INFO",
            "backupCount": backup_count,
            "encoding": "utf8",
        }
        config.loggers[name].handlers.append("file")

    logging.config.dictConfig(config)

    logger = logging.getLogger(name)
    for handler in logger.handlers:
        if handler.name != "file":
            continue
        if handler.suffix != "%Y-%m-%d.log":
            handler.suffix = "%Y-%m-%d.log"

    return logger
